/**
 * Module imprts
 */
var _       = require( 'underscore' )._;
var util    = require( 'util' );
var events  = require( 'events' );
var request = require( 'http' ).request;

/**
 * Module exports
 */
exports.Klient = Klient;

/**
 * Client Abstraction for HTTPServer/API tests
 *
 * @api public
 * @type {Function}
 * @param {options}
 * @return {Klient}
 */
function Klient ( options ) {
  events.EventEmitter.call( this );

  this.options = _.extend( {}, this.defaults, options );
  this.server = this.options.server;
  this.deferred = [];
  this.started = false;
  this.pending = false;
}

util.inherits( Klient, events.EventEmitter );

/**
 * Defaults
 *
 * @api private
 * @type {Object}
 * @param {String} host
 * @param {Number} port
 */
Klient.prototype.defaults = {
  host: 'localhost',
  port: 8081
};

/**
 * Handles queued requests
 *
 * @api private
 * @type {Function}
 */
Klient.prototype.processDeferred = function () {
  var self = this;
  var deferred = self.deferred;

  process.nextTick( function () {
    _( deferred.length ).times( function () {
      self.request( deferred.shift() );
    });
  });
};

/**
 * Lazily starts server
 *
 * @api private
 * @type {Function}
 */
Klient.prototype.startServer = function () {
  var self = this;
  var server = self.server;
  var host = self.options.host;
  var port = self.options.port;
  var callback;

  callback = function () {
    self.pending = false;
    self.started = true;
    self.processDeferred();
  };

  self.pending = true;

  server.listen( port, host, callback );
};

/**
 *
 *
 * @api public
 * @type {Function}
 * @param {Object} options
 */
Klient.prototype.request = function ( options ) {
  var self = this;
  var server = self.server;
  var promise = options.promise;
  var data, promise, req;

  if ( !server ) {
    return process.nextTick( function () {
      var err = new( Error)( 'Server is not defined' );

      promise.emit( 'error', err );
    });
  }

  if ( !server.fd || !self.started ) {
    self.deferred.push( options );

    if( !self.started && !self.pending ) {
      self.startServer();
    }

    return;
  }

  options = _.extend( {}, self.options, options );
  req = request(options);

  process.nextTick( function () {
    req.on( 'response', function ( res ) {
      var body = '';

      res.setEncoding('utf8');
      res.on('data', function(chunk) {
        body += chunk;
      });
      res.on('end', function() {
        if ( res.headers['content-type'] === 'application/json' ) {
          body = JSON.parse( body );
        }

        res.body = body;

        promise.emit( 'success', res );
      });
    }).on( 'error', function ( err ) {
      promise.emit( 'success', err );
    });

    if ( options.data ) {
      req.end( options.data );
    } else {
      req.end()
    }
  });
};

/**
 * Wrapper for request method to return a promise
 *
 * @api private
 * @type {Function}
 * @param {Object} options
 * @return {EventEmitter}
 */
Klient.prototype.promisedRequest = function ( options ) {
  var self = this;
  var promise = new( events.EventEmitter );

  options.promise = promise;

  self.request( options );

  return promise;
}

/**
 * GET convenience method
 *
 * @api public
 * @type {Function}
 * @return {EventEmitter}
 */
Klient.prototype.get = function ( options ) {
  _.extend( options, { method: 'get' } );

  return this.promisedRequest( options );
}

/**
 * POST convenience method
 *
 * @api public
 * @type {Function}
 * @return {EventEmitter}
 */
Klient.prototype.post = function ( options ) {
  _.extend( options, { method: 'post' } );

  return this.promisedRequest( options );
}

/**
 * PUT convenience method
 *
 * @api public
 * @type {Function}
 * @return {EventEmitter}
 */
Klient.prototype.put = function ( options ) {
  _.extend( options, { method: 'put' } );

  return this.promisedRequest( options );
}

/**
 * DELETE convenience method
 *
 * @api public
 * @type {Function}
 * @return {EventEmitter}
 */
Klient.prototype.del = function ( options ) {
  _.extend( options, { method: 'delete' } );

  return this.promisedRequest( options );
}

